(*----------------------------------------------------------------------------*
 *  Direct3D sample from DirectX 9.0 SDK Summer 2003                          *
 *  Delphi adaptation by Alexey Barkovoy (e-mail: clootie@ixbt.com)           *
 *                                                                            *
 *  Delphi versions 5-7 are supported                                         *
 *                                                                            *
 *  Modified: 08-Jan-2004                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru/delphi                                         *
 *----------------------------------------------------------------------------*)
//-----------------------------------------------------------------------------
// File: main.cpp
//
// Desc: Example code showing a technique for rendering volumetric fog in D3D.
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------

{$I DirectX.inc}

unit VolumeFogUnit;

interface

uses
  Windows, Messages, SysUtils, Math, CommDlg,
  Direct3D9, D3DX9, D3DApp, D3DEnumeration, D3DSettings,
  D3DFile, D3DFont, D3DUtil, DXUtil;

const
  RENDERTARGSIZE = 512;

// Maintains a set of triangles - its vertices and indices
type
  PTriangleSet = ^TTriangleSet;
  TTriangleSet = record
    pIndexBuffer: IDirect3DIndexBuffer9;
    pVertexBuffer: IDirect3DVertexBuffer9;
    dwNumVertices: DWORD;
    dwNumTriangles: DWORD;
    Center: TD3DXVector3;
    fRadius: Single;
  end;

  PFileVertex = ^TFileVertex;
  TFileVertex = packed record
    x,y,z: Single;
    nx,ny,nz: Single;
  end;

const
  TFileVertex_FVF = (D3DFVF_XYZ or D3DFVF_NORMAL);

type
  //-----------------------------------------------------------------------------
  // Name: class CMyD3DApplication
  // Desc: Application class. The base class (CD3DApplication) provides the
  //       generic functionality needed in all Direct3D samples. CMyD3DApplication
  //       adds functionality specific to this sample program.
  //-----------------------------------------------------------------------------
  CMyD3DApplication = class(CD3DApplication)
  protected
    m_matObject:        TD3DXMatrix;
    m_matView:          TD3DXMatrix;
    m_matProj:          TD3DXMatrix;
    m_matShadObject:    TD3DXMatrix;

//    m_bInsideFog:       Boolean; //BAA: never used anyway
    m_bUsePicShader:    Boolean; 
    m_bAutoMove:        Boolean;
    m_bShowHelp:        Boolean;

    m_pScreenVertBuf:   IDirect3DVertexBuffer9;
    m_pScreenVertBuf2:  IDirect3DVertexBuffer9;
    m_pOldRenderTarget: IDirect3DSurface9;
    m_pOldStencilZ:     IDirect3DSurface9;
    m_pTargetZ:         IDirect3DSurface9;

    m_FogVolume:        TTriangleSet;
    m_MeshShaderObject: TTriangleSet;

    m_bKey:             array [0..255] of Boolean; // Byte;
    m_vVelocity:        TD3DXVector3;
    m_fYawVelocity:     Single;
    m_fPitchVelocity:   Single;
    m_matOrientation:   TD3DXMatrix;
    m_vPosition:        TD3DXVector3;
    m_fYaw:             Single;
    m_fPitch:           Single;

    m_pBackFogDist:     IDirect3DTexture9;
    m_pFrontFogDist:    IDirect3DTexture9;
    m_pSceneDist:       IDirect3DTexture9;
    m_pTempDist:        IDirect3DTexture9;

    m_pStepTexture:     IDirect3DTexture9;

    m_pFogDeclaration:  IDirect3DVertexDeclaration9;
    m_pFogShader:       IDirect3DVertexShader9;
    m_pFogDeclaration2: IDirect3DVertexDeclaration9;
    m_pFogShader2:      IDirect3DVertexShader9;
    m_pSubPicShader:    IDirect3DPixelShader9;

    m_pMeshObject:      ID3DXMesh;

    m_pFont:            CD3DFont;
    m_pFontSmall:       CD3DFont;
  protected
    function InitDeviceObjects: HResult; override;
    function RestoreDeviceObjects: HResult; override;
    function InvalidateDeviceObjects: HResult; override;
    function DeleteDeviceObjects: HResult; override;
    function FinalCleanup: HResult; override;
    function FrameMove: HResult; override;
    function Render: HResult; override;
    function ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
      adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT; override;
    /////
    function SetFogMatrices: HRESULT;
    function SetObjMatrices: HRESULT;
    function LoadXFile(pModel: PChar; pTriSet: PTriangleSet): HRESULT;
    function ComputeStepTexture: HRESULT;
  public
    constructor Create;
    destructor Destroy; override;
    function MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT; override;
 end;

implementation


{ CMyD3DApplication }

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
constructor CMyD3DApplication.Create;
var
  i: Integer;
begin
  inherited;

  m_d3dEnumeration.AppUsesDepthBuffer := True;

  m_strWindowTitle  := 'VolumeFog';

  m_pFont           := CD3DFont.Create('Arial', 12, D3DFONT_BOLD);
  m_pFontSmall      := CD3DFont.Create('Arial',  9, D3DFONT_BOLD);

  for i:= 0 to 255 do m_bKey[i]:= False;

  D3DXMatrixIdentity(m_matObject);
  D3DXMatrixIdentity(m_matView);
  D3DXMatrixIdentity(m_matProj);
  D3DXMatrixIdentity(m_matShadObject);
  D3DXMatrixIdentity(m_matOrientation);

  m_vPosition      := D3DXVector3(0.0, 0.0,-4.0);
  m_vVelocity      := D3DXVector3(0.0, 0.0, 0.0);
  m_fYaw           := 0.0;
  m_fYawVelocity   := 0.0;
  m_fPitch         := 0.0;
  m_fPitchVelocity := 0.0;
  m_bShowHelp      := False;
  m_bAutoMove      := True;

  m_pOldRenderTarget := nil;
  m_pOldStencilZ     := nil;
  m_pScreenVertBuf   := nil;
  m_pBackFogDist     := nil;
  m_pSceneDist       := nil;
  m_pTempDist        := nil;
  m_pFrontFogDist    := nil;
  m_pTargetZ         := nil;
  m_pMeshObject      := nil;
  m_pStepTexture     := nil;
  m_pScreenVertBuf   := nil;
  m_pScreenVertBuf2  := nil;
  m_pFogDeclaration  := nil;
  m_pFogShader       := nil;
  m_pFogDeclaration2 := nil;
  m_pFogShader2      := nil;
  m_pSubPicShader    := nil;

  FillChar(m_MeshShaderObject, SizeOf(m_MeshShaderObject), 0);
  FillChar(m_FogVolume, SizeOf(m_FogVolume), 0);
end;

destructor CMyD3DApplication.Destroy;
begin
  inherited;
end;


//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device initialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
function CMyD3DApplication.ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
  adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT;
begin
  Result:= E_FAIL;

  if ((dwBehavior and D3DCREATE_HARDWARE_VERTEXPROCESSING <> 0) or
      (dwBehavior and D3DCREATE_MIXED_VERTEXPROCESSING <> 0)) then
  begin
    if (pCaps.VertexShaderVersion < D3DVS_VERSION(1,0)) then Exit;
  end;

  if (pCaps.PixelShaderVersion < D3DPS_VERSION(1,1)) then Exit;

  if (dwBehavior and D3DCREATE_PUREDEVICE <> 0) then Exit;

  if not (backBufferFormat in [D3DFMT_A8R8G8B8, D3DFMT_X8R8G8B8, D3DFMT_R8G8B8]) then Exit;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Process windows messages.
//-----------------------------------------------------------------------------
function CMyD3DApplication.MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM;
  lParam: LPARAM): LRESULT;
begin
  case uMsg of
    WM_KEYUP:
    begin
      m_bKey[wParam] := False;
      if (wParam = VK_F1) then m_bShowHelp := not m_bShowHelp;
    end;

    WM_KEYDOWN:
      m_bKey[wParam] := True;
  end;

  Result := inherited MsgProc(hWnd, uMsg, wParam, lParam);
end;


//-----------------------------------------------------------------------------
// Name: ComputeStepTexture()
// Desc: This computes a step texture - used for high precision fogging.  The
//       step texture is a 4096 texture which increments from 0 to 4095. This
//       is used because on a triangle basis - the texture interpolators are
//       much higher precision than the color interpolators. Since we need a
//       higher precision then 8 bits, we need a way to load in more bits into
//       every pixel. This is only way to do it on an 8 bit card.
//-----------------------------------------------------------------------------
function CMyD3DApplication.ComputeStepTexture: HRESULT;
var
  Lock: TD3DLockedRect;
  Width: Cardinal;
  Height: Cardinal;
  pLine, pBase: PByteArray;
  x, y: Cardinal;
  widthRatio: Single;
  scaledX: Cardinal;

  desc: TD3DSurfaceDesc;
begin
  Width  := 4096;
  Height := 8;
  widthRatio := 1.0;

  Result:= D3DXCreateTexture(m_pd3dDevice, Width, Height, 1, 0, D3DFMT_A8R8G8B8,
                             D3DPOOL_MANAGED, m_pStepTexture);
  if FAILED(Result) then Exit;

  // Adjust Width if D3DXCreateTexture had to pick a different texture
  // width.  Texture contents will still go up to 4096, though.
  m_pStepTexture.GetLevelDesc(0, desc);
  if (desc.Width <> Width) then
  begin
    widthRatio := Width / desc.Width;
    Width := desc.Width;
  end;

  Result:= m_pStepTexture.LockRect(0, Lock, nil, 0);
  if FAILED(Result) then
  begin
    m_pStepTexture:= nil;
    Exit;
  end;

  pBase := Lock.pBits;

  for y:= 0 to Height - 1 do
  begin
    pLine := @pBase[y*Cardinal(Lock.Pitch)];
    for x:= 0 to Width - 1 do
    begin
      scaledX:= Trunc(widthRatio * x);
      pLine[0] :=  scaledX and $F;  // red, lower 4 bits
      pLine[2] := (scaledX and $FF0) shr 4; // blue, upper 8 bits
      pLine[1] := 0; // green
      pLine[3] := 255; // alpha
      Inc(Integer(pLine), 4);
    end;
  end;

  m_pStepTexture.UnlockRect(0);

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: LoadXFile()
// Desc: This loads an xfile into the TRIANGLESET structure.  It is not enough
//       to render an X-File blindly since the the data must be fed into our
//       vertex shader.
//-----------------------------------------------------------------------------
function CMyD3DApplication.LoadXFile(pModel: PChar; pTriSet: PTriangleSet): HRESULT;
var
  pbufMaterials: ID3DXBuffer;
  pbufAdjacency: ID3DXBuffer;
  pVertexBuffer: IDirect3DVertexBuffer9;
  pIndexBuffer: IDirect3DIndexBuffer9;
  IndexDesc: TD3DIndexBufferDesc;
  cMaterials: DWORD;
  pSourceI: Pointer;
  pDestI: Pointer;
  pMeshSysMem: ID3DXMesh;
  pTempMesh: ID3DXMesh;
  pVert: PFileVertex;
  pSource: PFileVertex;
  dwGetFVF: DWORD;
begin
  pTriSet.pVertexBuffer := nil;
  pTriSet.pIndexBuffer  := nil;
  pVertexBuffer := nil;

  Result:= S_OK;
  try
    //load the x file
    Result:= D3DXLoadMeshFromX(pModel, D3DXMESH_SYSTEMMEM, m_pd3dDevice,
                               @pbufAdjacency, @pbufMaterials, nil, @cMaterials, pTempMesh);
    if FAILED(Result) then Exit;

    dwGetFVF := pTempMesh.GetFVF;

    // Change it to a mesh that has this vertex format if it is not that way
    if (dwGetFVF <>  TFileVertex_FVF) then
    begin
      Result:= pTempMesh.CloneMeshFVF(0, TFileVertex_FVF, m_pd3dDevice, pMeshSysMem);
      pTempMesh:= nil;
      if FAILED(Result) then Exit;
    end else
    begin
      pMeshSysMem := pTempMesh;
    end;

    pTriSet.dwNumTriangles := pMeshSysMem.GetNumFaces;
    pTriSet.dwNumVertices  := pMeshSysMem.GetNumVertices;

    // Now copy the vertex buffers and index buffers into new ones that
    // have things set the way we want

    Result:= pMeshSysMem.GetVertexBuffer(pVertexBuffer);
    if FAILED(Result) then Exit;

    Result:= m_pd3dDevice.CreateVertexBuffer(SizeOf(TFileVertex) * pTriSet.dwNumVertices,
        D3DUSAGE_WRITEONLY, TFileVertex_FVF, D3DPOOL_MANAGED,
        pTriSet.pVertexBuffer, nil);
    if FAILED(Result) then Exit;

    Result:= pTriSet.pVertexBuffer.Lock(0, 0, Pointer(pVert), 0);
    if FAILED(Result) then Exit;

    Result:= pVertexBuffer.Lock(0, SizeOf(TFileVertex)*pTriSet.dwNumVertices, Pointer(pSource), 0);
    if FAILED(Result) then Exit;

    CopyMemory(pVert, pSource, SizeOf(TFileVertex) * pTriSet.dwNumVertices);

    D3DXComputeBoundingSphere(PD3DXVector3(pVert), pTriSet.dwNumVertices,
      D3DXGetFVFVertexSize(TFileVertex_FVF), pTriSet.Center, pTriSet.fRadius);

    pVertexBuffer.Unlock;
    pTriSet.pVertexBuffer.Unlock;

    Result:= pMeshSysMem.GetIndexBuffer(pIndexBuffer);
    if FAILED(Result) then Exit;

    pIndexBuffer.GetDesc(IndexDesc);

    Result:= m_pd3dDevice.CreateIndexBuffer(IndexDesc.Size, D3DUSAGE_WRITEONLY,
        IndexDesc.Format, D3DPOOL_MANAGED, pTriSet.pIndexBuffer, nil);
    if FAILED(Result) then Exit;

    pIndexBuffer.Lock(0, IndexDesc.Size, pSourceI, 0);
    pTriSet.pIndexBuffer.Lock(0, IndexDesc.Size, pDestI, 0);
    CopyMemory(pDestI, pSourceI, IndexDesc.Size);

    pIndexBuffer.Unlock;
    pTriSet.pIndexBuffer.Unlock;

    //cleanup
    SAFE_RELEASE(pbufMaterials);
    SAFE_RELEASE(pbufAdjacency);
    SAFE_RELEASE(pMeshSysMem);
    SAFE_RELEASE(pVertexBuffer);
    SAFE_RELEASE(pIndexBuffer);

    Result:= S_OK;

  finally
    if FAILED(Result) then
    begin
      SAFE_RELEASE(pMeshSysMem);
      SAFE_RELEASE(pVertexBuffer);
      SAFE_RELEASE(pTriSet.pVertexBuffer);
      SAFE_RELEASE(pTriSet.pIndexBuffer);
      SAFE_RELEASE(pIndexBuffer);
      SAFE_RELEASE(pbufMaterials);
      SAFE_RELEASE(pbufAdjacency);
    end;
  end;
end;


//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FrameMove: HResult;
var
  fElapsedTime: Single;
  fSpeed: Single;
  fAngularSpeed: Single;
  vT: TD3DXVector3;
  qR: TD3DXQuaternion;
begin
  if (m_fElapsedTime > 0.0)
  then fElapsedTime := m_fElapsedTime
  else fElapsedTime := 0.05;

  fSpeed        := 6.0*fElapsedTime;
  fAngularSpeed := 1.0*fElapsedTime;

  // De-accelerate the camera movement (for smooth motion)
  D3DXVec3Scale(m_vVelocity, m_vVelocity, 0.9);
  m_fYawVelocity   := m_fYawVelocity*0.9;
  m_fPitchVelocity := m_fPitchVelocity*0.9;

  // Process keyboard input
  if (m_bKey[VK_RIGHT])  then m_vVelocity.x    := m_vVelocity.x + fSpeed; // Slide Right
  if (m_bKey[VK_LEFT])   then m_vVelocity.x    := m_vVelocity.x - fSpeed; // Slide Left
  if (m_bKey[VK_UP])     then m_vVelocity.y    := m_vVelocity.y + fSpeed; // Slide Up
  if (m_bKey[VK_DOWN])   then m_vVelocity.y    := m_vVelocity.y - fSpeed; // Slide Down
  if (m_bKey[Byte('W')]) then m_vVelocity.z    := m_vVelocity.z + fSpeed; // Move Forward
  if (m_bKey[Byte('S')]) then m_vVelocity.z    := m_vVelocity.z - fSpeed; // Move Backward
  if (m_bKey[Byte('E')]) then m_fYawVelocity   := m_fYawVelocity + fSpeed; // Turn Right
  if (m_bKey[Byte('Q')]) then m_fYawVelocity   := m_fYawVelocity - fSpeed; // Turn Left
  if (m_bKey[Byte('Z')]) then m_fPitchVelocity := m_fPitchVelocity + fSpeed; // Turn Down
  if (m_bKey[Byte('A')]) then m_fPitchVelocity := m_fPitchVelocity - fSpeed; // Turn Up

  // Stop the camera if the camera is moved by hand
  if (m_vVelocity.x <> 0.0) or (m_vVelocity.y <> 0.0) or
     (m_fYawVelocity <> 0.0) or (m_fPitchVelocity <> 0.0)
  then m_bAutoMove := False;

  // Move the camera back and forth
  if m_bAutoMove then m_vVelocity.z := Cos(m_fTime)*0.5;

  // Update the position vector
  D3DXVec3Scale(vT, m_vVelocity, fSpeed);
  D3DXVec3TransformNormal(vT, vT, m_matOrientation);
  D3DXVec3Add(m_vPosition, m_vPosition, vT);

  // Update the yaw-pitch-rotation vector
  m_fYaw   := m_fYaw + fAngularSpeed * m_fYawVelocity;
  m_fPitch := m_fPitch + fAngularSpeed * m_fPitchVelocity;

  // Set the view matrix
  D3DXQuaternionRotationYawPitchRoll(qR, m_fYaw, m_fPitch, 0.0);
  D3DXMatrixAffineTransformation(m_matOrientation, 1.25, nil, @qR, @m_vPosition);
  D3DXMatrixInverse(m_matView, nil, m_matOrientation);
  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: SetFogMatrices()
// Desc: Sets up the fog matrices, the matrices which represent the fog volume's
//       position.  These have to be loaded into the right places in the vertex
//       shader:
//       0-3 = object transformation matrix
//       16-19 = ObjectView
//       8-11 = Total matrix
//-----------------------------------------------------------------------------
function CMyD3DApplication.SetFogMatrices: HRESULT;
var
  temp, tObject, m_matTot: TD3DXMatrix;
begin
  D3DXMatrixMultiply(m_matTot, m_matObject, m_matView);
  D3DXMatrixTranspose(temp, m_matTot);
  m_pd3dDevice.SetVertexShaderConstantF(16, @temp, 4);

  D3DXMatrixMultiply(m_matTot, m_matTot, m_matProj);

  D3DXMatrixTranspose(m_matTot, m_matTot);
  D3DXMatrixTranspose(tObject, m_matObject);

  //these need to be transposes
  m_pd3dDevice.SetVertexShaderConstantF(0, @tObject, 4);
  m_pd3dDevice.SetVertexShaderConstantF(8, @m_matTot, 4);

  //the w clip distance

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: SetObjMatrices()
// Desc: Sets up the object's matrices, the matrices which represent the object
//       (or the world) volume's position.  These have to be loaded into the
//       right places in the vertex shader:
//       0-3 = object transformation matrix
//       16-19 = ObjectView
//-----------------------------------------------------------------------------
function CMyD3DApplication.SetObjMatrices: HRESULT;
var
  temp, tObject, m_matTot: TD3DXMatrix;
begin
  D3DXMatrixTranspose(tObject, m_matShadObject);
  m_pd3dDevice.SetVertexShaderConstantF(0, @tObject, 4);

  D3DXMatrixMultiply(m_matTot, m_matShadObject, m_matView);
  D3DXMatrixTranspose(temp, m_matTot);
  m_pd3dDevice.SetVertexShaderConstantF(16, @temp, 4);

  D3DXMatrixMultiply(m_matTot, m_matTot, m_matProj);

  D3DXMatrixTranspose(m_matTot, m_matTot);
  m_pd3dDevice.SetVertexShaderConstantF(8, @m_matTot, 4);

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.Render: HResult;
var
  lightDir: TD3DXVector3;
  WClip: TD3DXVector4;

  pSurface: IDirect3DSurface9;
  Mat: TD3DMaterial9;

  ViewPort, OldPort: TD3DViewport9;
  myLight: TD3DLight9;

  fAspectRatio: Single;
  vDesc: TD3DVertexBufferDesc;
begin
  lightDir:= D3DXVector3(1,0,0);

  //WClip represents our near and far fog 'clipping' (the y index is the inverse of the rear
  //clipping plane)
  //planes. The higher precision one has a further clipping plane

  WClip := D3DXVector4(1,0.015,1,1);

  // Compute the current viewport
  // Since we might be on a pure device, we can't query it from the driver
  OldPort.X := 0;
  OldPort.Y := 0;
  OldPort.Width := m_d3dsdBackBuffer.Width;
  OldPort.Height := m_d3dsdBackBuffer.Height;
  OldPort.MaxZ := 1;
  OldPort.MinZ := 0;

  fAspectRatio := OldPort.Width / OldPort.Height;

  // Update projection matrix for a changed aspect ratio
  D3DXMatrixPerspectiveFovLH(m_matProj, 0.75, fAspectRatio, 1.0, 100);

  lightDir := D3DXVector3(2.5, 2.25, -2.55);

  //set the fog clipping into the vertex shader's constant register #12
  m_pd3dDevice.SetVertexShaderConstantF(12, @WClip, 1);

  m_pd3dDevice.BeginScene;

  //now we render the first half of the fog scene - or rather, the back side of the
  //fog. what will be left in the buffer after we are done is the w depth of the
  //back fog volume

  ViewPort.X := 0;
  ViewPort.Y := 0;
  ViewPort.Height := RENDERTARGSIZE;
  ViewPort.Width  := RENDERTARGSIZE;
  ViewPort.MaxZ := 1;
  ViewPort.MinZ := 0;

  m_pd3dDevice.SetTexture(0, nil);
  m_pd3dDevice.SetTexture(1, nil);
  m_pd3dDevice.SetTexture(2, nil);
  m_pd3dDevice.SetTexture(3, nil);

  m_pd3dDevice.SetPixelShader(nil);

  //we will just use a special vertex shader to fill the color and alpha as our w depth using the
  //step texture
  m_pd3dDevice.SetRenderState(D3DRS_WRAP0, 0);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_TEXCOORDINDEX, 0);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_SELECTARG1);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_SELECTARG1);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TFACTOR);
  m_pd3dDevice.SetRenderState(D3DRS_TEXTUREFACTOR, $01010101);

  m_pd3dDevice.SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_POINT);
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_POINT);

  //use our step texture to get higher precision interpolating
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_MIPFILTER, D3DTEXF_NONE);

  m_pd3dDevice.SetTextureStageState(1, D3DTSS_COLOROP, D3DTOP_DISABLE);
  m_pd3dDevice.SetTextureStageState(1, D3DTSS_ALPHAOP, D3DTOP_DISABLE);

  m_pd3dDevice.SetSamplerState(0, D3DSAMP_ADDRESSU, D3DTADDRESS_CLAMP);
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_ADDRESSV, D3DTADDRESS_CLAMP);

  m_pd3dDevice.SetTexture(0, m_pStepTexture);
  m_pd3dDevice.SetTexture(1, nil);

  //set up transform matrices for the fixed pipe.
  //this is only used for the rendering of the object
  m_pd3dDevice.SetTransform(D3DTS_VIEW, m_matView);
  m_pd3dDevice.SetTransform(D3DTS_PROJECTION, m_matProj);
  m_pd3dDevice.SetTransform(D3DTS_WORLD, m_matObject);

  //Now we need to draw the scene, happens first to also set the Z
  m_pSceneDist.GetSurfaceLevel(0, pSurface);

  //first, render the scene with the depth encoding
  m_pd3dDevice.SetRenderTarget(0, pSurface);
  m_pd3dDevice.SetDepthStencilSurface(m_pTargetZ);
  m_pd3dDevice.SetViewport(ViewPort);
  m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER ,
                     D3DCOLOR_ARGB($00, $00, $00, $00), 1.0, 0);

  m_pd3dDevice.SetRenderState(D3DRS_CULLMODE, D3DCULL_CCW);

  SetObjMatrices;
  m_pd3dDevice.SetRenderState(D3DRS_TEXTUREFACTOR, $00000000);
  m_pd3dDevice.SetIndices(m_MeshShaderObject.pIndexBuffer);
  m_pd3dDevice.SetStreamSource(0, m_MeshShaderObject.pVertexBuffer, 0, SizeOf(TFileVertex));
  m_pd3dDevice.SetVertexDeclaration(m_pFogDeclaration);
  m_pd3dDevice.SetVertexShader(m_pFogShader);

  //no alpha blending since geometry is opaque
  m_pd3dDevice.SetRenderState(D3DRS_ZWRITEENABLE, iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_ZENABLE, iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_ALPHABLENDENABLE, iFalse);
  m_pd3dDevice.DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, 0,
      m_MeshShaderObject.dwNumVertices, 0, m_MeshShaderObject.dwNumTriangles);

  //Now Render the back side of the fog, the
  //z buffer is the same as the one set by the geometry
  //since elements of the fog behind the scene should not be rendered (since they are not visible)
  pSurface:= nil;
  m_pBackFogDist.GetSurfaceLevel(0, pSurface);
  m_pd3dDevice.SetRenderTarget(0, pSurface);
  m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET, D3DCOLOR_ARGB($00, $00, $00, $00), 1.0, 0);

  SetFogMatrices;

  //reverse the culling order to get the back side
  m_pd3dDevice.SetRenderState(D3DRS_CULLMODE, D3DCULL_CW);

  //add up all the backsides, this is volumetric rendering so
  //everything is totally transparent, want to check Z Buffer,
  //but not write to it.
  m_pd3dDevice.SetRenderState(D3DRS_ALPHABLENDENABLE, iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_ALPHATESTENABLE, iFalse);
  m_pd3dDevice.SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ONE);
  m_pd3dDevice.SetRenderState(D3DRS_SRCBLEND, D3DBLEND_ONE);
  m_pd3dDevice.SetRenderState(D3DRS_ZWRITEENABLE, iFalse);

  //the tfactor adds in the alpha so that we can keep track of the number
  //of times a pixel is in and out of a fog
  m_pd3dDevice.SetRenderState(D3DRS_TEXTUREFACTOR, $10101010);
  m_pd3dDevice.SetIndices(m_FogVolume.pIndexBuffer);
  m_pd3dDevice.SetStreamSource(0, m_FogVolume.pVertexBuffer, 0, SizeOf(TFileVertex));
  m_pd3dDevice.SetVertexDeclaration(m_pFogDeclaration);
  m_pd3dDevice.SetVertexShader(m_pFogShader);
  m_pd3dDevice.DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, 0,
      m_FogVolume.dwNumVertices, 0, m_FogVolume.dwNumTriangles);
  pSurface:= nil;

  //Now we need to draw the front end of the fog
  //same concept as above, but this time the front halves of the
  //fog get drawn
  m_pFrontFogDist.GetSurfaceLevel(0, pSurface);
  m_pd3dDevice.SetRenderTarget(0, pSurface);
  m_pd3dDevice.SetViewport(ViewPort);
  m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET,
                     D3DCOLOR_ARGB($00, $00, $00, $00), 1.0, 0);

  m_pd3dDevice.SetRenderState(D3DRS_ALPHABLENDENABLE, iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_ALPHATESTENABLE, iFalse);
  m_pd3dDevice.SetRenderState(D3DRS_TEXTUREFACTOR, $10101010);
  m_pd3dDevice.SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ONE);//DESTALPHA);
  m_pd3dDevice.SetRenderState(D3DRS_SRCBLEND, D3DBLEND_ONE);
  m_pd3dDevice.SetRenderState(D3DRS_ZWRITEENABLE, iFalse);
  m_pd3dDevice.SetRenderState(D3DRS_CULLMODE, D3DCULL_CCW);

  SetFogMatrices;
  m_pd3dDevice.SetIndices(m_FogVolume.pIndexBuffer);
  m_pd3dDevice.SetStreamSource(0, m_FogVolume.pVertexBuffer, 0, SizeOf(TFileVertex));
  m_pd3dDevice.SetVertexDeclaration(m_pFogDeclaration);
  m_pd3dDevice.SetVertexShader(m_pFogShader);
  m_pd3dDevice.DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, 0,
      m_FogVolume.dwNumVertices, 0, m_FogVolume.dwNumTriangles);
  pSurface:= nil;

  //first we render the object, or whatever solid objects
  //we have in the scene. This will act as the back side of our
  //fog if it happens to be in the fog volume

  m_pd3dDevice.SetRenderState(D3DRS_ALPHABLENDENABLE, iFalse);
  m_pd3dDevice.SetRenderState(D3DRS_ALPHATESTENABLE, iFalse);

  m_pd3dDevice.SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ONE);
  m_pd3dDevice.SetRenderState(D3DRS_SRCBLEND, D3DBLEND_ONE);
  m_pd3dDevice.SetRenderState(D3DRS_ZWRITEENABLE, iTrue);

  //now we are ready to actually draw the scene
  m_pd3dDevice.SetRenderTarget(0, m_pOldRenderTarget);
  m_pd3dDevice.SetDepthStencilSurface(m_pOldStencilZ);
  m_pd3dDevice.SetViewport(OldPort);

  m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER,
                     D3DCOLOR_ARGB($00, $00, $00, $00), 1.0, 0);

  //first draw the scene as it is.
  m_pd3dDevice.SetTransform(D3DTS_VIEW, m_matView);
  m_pd3dDevice.SetTransform(D3DTS_PROJECTION, m_matProj);
  m_pd3dDevice.SetTransform(D3DTS_WORLD, m_matShadObject);
  m_pd3dDevice.SetRenderState(D3DRS_CULLMODE, D3DCULL_CCW);

  FillChar(Mat, SizeOf(Mat), 0);

  //generic light and generic scene.
  FillChar(myLight, SizeOf(myLight), 0); //TODO: Watch when Microsoft will fix this bug 
  myLight._Type := D3DLIGHT_DIRECTIONAL;
  myLight.Diffuse.r := 0;
  myLight.Diffuse.g := 1;
  myLight.Diffuse.b := 0;
  myLight.Diffuse.a := 1;
  myLight.Range     := 1000000.0;
  myLight.Direction := D3DXVector3(1,1,1);

  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_SELECTARG1);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_SELECTARG1);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_DIFFUSE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_DIFFUSE);

  m_pd3dDevice.SetTextureStageState(1, D3DTSS_COLOROP, D3DTOP_DISABLE);
  m_pd3dDevice.SetTextureStageState(1, D3DTSS_ALPHAOP, D3DTOP_DISABLE);

  m_pd3dDevice.SetRenderState(D3DRS_LIGHTING, iTrue);
  m_pd3dDevice.LightEnable(0, True);

  m_pd3dDevice.SetLight(0, myLight);

  Mat.Diffuse.r := 1;
  Mat.Diffuse.b := 1;
  Mat.Diffuse.g := 1;
  Mat.Diffuse.a := 1;

  m_pd3dDevice.SetMaterial(Mat);

  m_pd3dDevice.SetTexture(0, nil);
  m_pd3dDevice.SetTexture(1, nil);

  m_MeshShaderObject.pVertexBuffer.GetDesc(vDesc);
  m_pd3dDevice.SetStreamSource(0, m_MeshShaderObject.pVertexBuffer, 0, SizeOf(TFileVertex));
  m_pd3dDevice.SetIndices(m_MeshShaderObject.pIndexBuffer);
  m_pd3dDevice.SetFVF(TFileVertex_FVF);
  m_pd3dDevice.SetVertexShader(nil);
  m_pd3dDevice.DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, 0,
    m_MeshShaderObject.dwNumVertices, 0, m_MeshShaderObject.dwNumTriangles);

  m_pd3dDevice.SetTexture(0, nil);
  m_pd3dDevice.SetTexture(1, nil);
  m_pd3dDevice.SetTexture(2, nil);
  m_pd3dDevice.SetTexture(3, nil);

  //draw the fog
  //this uses a complex pixel shader which is fully commented
  //this computes the amount of fog at every pixel on the screen

  m_pd3dDevice.SetRenderTarget(0, m_pOldRenderTarget);
  m_pd3dDevice.SetDepthStencilSurface(m_pOldStencilZ);
  m_pd3dDevice.SetViewport(OldPort);

  m_pd3dDevice.SetRenderState(D3DRS_ZENABLE, iFalse);
  m_pd3dDevice.SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);

  m_pd3dDevice.SetRenderState(D3DRS_WRAP0, 0);
  m_pd3dDevice.SetRenderState(D3DRS_WRAP1, 0);
  m_pd3dDevice.SetRenderState(D3DRS_WRAP2, 0);

  m_pd3dDevice.SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_POINT);
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_POINT);
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_MIPFILTER, D3DTEXF_NONE);
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_ADDRESSU,D3DTADDRESS_CLAMP);
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_ADDRESSV,D3DTADDRESS_CLAMP);

  m_pd3dDevice.SetSamplerState(1, D3DSAMP_MINFILTER, D3DTEXF_POINT);
  m_pd3dDevice.SetSamplerState(1, D3DSAMP_MAGFILTER, D3DTEXF_POINT);
  m_pd3dDevice.SetSamplerState(1, D3DSAMP_MIPFILTER, D3DTEXF_NONE);
  m_pd3dDevice.SetSamplerState(1, D3DSAMP_ADDRESSU,D3DTADDRESS_CLAMP);
  m_pd3dDevice.SetSamplerState(1, D3DSAMP_ADDRESSV,D3DTADDRESS_CLAMP);

  m_pd3dDevice.SetSamplerState(2, D3DSAMP_MINFILTER, D3DTEXF_POINT);
  m_pd3dDevice.SetSamplerState(2, D3DSAMP_MAGFILTER, D3DTEXF_POINT);
  m_pd3dDevice.SetSamplerState(2, D3DSAMP_MIPFILTER, D3DTEXF_NONE);
  m_pd3dDevice.SetSamplerState(2, D3DSAMP_ADDRESSU,D3DTADDRESS_CLAMP);
  m_pd3dDevice.SetSamplerState(2, D3DSAMP_ADDRESSV,D3DTADDRESS_CLAMP);

  m_pd3dDevice.SetTextureStageState(1, D3DTSS_TEXCOORDINDEX, 0);
  m_pd3dDevice.SetTextureStageState(2, D3DTSS_TEXCOORDINDEX, 0);

  m_pd3dDevice.SetPixelShader(m_pSubPicShader);
  m_pd3dDevice.SetTexture(0, m_pFrontFogDist);
  m_pd3dDevice.SetTexture(1, m_pBackFogDist);
  m_pd3dDevice.SetTexture(2, m_pSceneDist);

  //alpha blend on. If we want to do volumetric light, we use an additive rather
  //then blended mode
  m_pd3dDevice.SetRenderState(D3DRS_ALPHABLENDENABLE, iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA);
  m_pd3dDevice.SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);


  m_pd3dDevice.SetTextureStageState(1, D3DTSS_TEXCOORDINDEX, 1);
  m_pd3dDevice.SetTextureStageState(2, D3DTSS_TEXCOORDINDEX, 2);

  //simple vertex shader, just copies coordinates and stuff.
  m_pd3dDevice.SetVertexDeclaration(m_pFogDeclaration2);
  m_pd3dDevice.SetVertexShader(m_pFogShader2);
  m_pd3dDevice.SetStreamSource(0, m_pScreenVertBuf, 0, 7 * SizeOf(Single));
  m_pd3dDevice.DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);

  m_pd3dDevice.SetRenderState(D3DRS_ALPHABLENDENABLE, iFalse);

  // Output statistics
  m_pFont.DrawText(2,  0, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats);
  m_pFont.DrawText(2, 20, D3DCOLOR_ARGB(255,255,255,0), m_strDeviceStats);

  // Draw help
  if m_bShowHelp then
  begin
    m_pFontSmall.DrawText(2, 40, $ffffff00, 'Keyboard controls:');
    m_pFontSmall.DrawText(20, 60, $ffffff00,
                          'Move'#10'Turn'#10'Pitch'#10'Slide'#10#10'Help'#10+
                          'Change device'#10+
                          'Exit');
    m_pFontSmall.DrawText(160, 60, $ffffff00,
                          'W,S'#10'E,Q'#10'A,Z'#10'Arrow keys'#10#10+
                          'F1'#10'F2'#10'Esc');
  end else
  begin
    m_pFontSmall.DrawText(2, 40, $ffffff00, 'Press F1 for help');
  end;

  m_pd3dDevice.EndScene;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
var
  declVS: array[0..2] of TD3DVertexElement9 =
  (
    (Stream: 0; Offset: 0; _Type: D3DDECLTYPE_FLOAT3; Method: D3DDECLMETHOD_DEFAULT; Usage: D3DDECLUSAGE_POSITION; UsageIndex: 0),
    (Stream: 0; Offset: 12; _Type: D3DDECLTYPE_FLOAT3; Method: D3DDECLMETHOD_DEFAULT; Usage: D3DDECLUSAGE_NORMAL; UsageIndex: 0),
    ({D3DDECL_END})
  );
  declVS2: array[0..3] of TD3DVertexElement9 =
  (
    (Stream: 0; Offset: 0; _Type: D3DDECLTYPE_FLOAT4; Method: D3DDECLMETHOD_DEFAULT; Usage: D3DDECLUSAGE_POSITION; UsageIndex: 0),
    (Stream: 0; Offset: 16; _Type: D3DDECLTYPE_D3DCOLOR; Method: D3DDECLMETHOD_DEFAULT; Usage: D3DDECLUSAGE_COLOR; UsageIndex: 0),
    (Stream: 0; Offset: 20; _Type: D3DDECLTYPE_FLOAT2; Method: D3DDECLMETHOD_DEFAULT; Usage: D3DDECLUSAGE_TEXCOORD; UsageIndex: 0),
    ({D3DDECL_END})
  );
function CMyD3DApplication.InitDeviceObjects: HResult;
var
  model: array[0..159] of Char;
  model2: array[0..159] of Char;
  strPath: array[0..MAX_PATH-1] of Char;
  pCodeShader: ID3DXBuffer;
begin
  m_pFont.InitDeviceObjects(m_pd3dDevice);
  m_pFontSmall.InitDeviceObjects(m_pd3dDevice);

  Result:= D3DAPPERR_MEDIANOTFOUND;

  if FAILED(DXUtil_FindMediaFileCb(model, SizeOf(model), 'fog2.x')) then Exit;
  if FAILED(DXUtil_FindMediaFileCb( model2, SizeOf(model2), 'room2.x')) then Exit;

  m_bUsePicShader:= (D3DSHADER_VERSION_MAJOR(m_d3dCaps.PixelShaderVersion) >= 1);

  Result:= LoadXFile(model, @m_FogVolume);
  if FAILED(Result) then Exit;

  Result:= LoadXFile(model2, @m_MeshShaderObject);
  if FAILED(Result) then Exit;

  Result:= D3DXLoadMeshFromX(model2, D3DXMESH_SYSTEMMEM, m_pd3dDevice,
      nil, nil, nil, nil, m_pMeshObject);
  if FAILED(Result) then Exit;

  Result:= m_pd3dDevice.CreateVertexBuffer(7 * SizeOf(Single) * 4, 0,
      D3DFVF_XYZRHW or D3DFVF_TEX1 or D3DFVF_DIFFUSE, D3DPOOL_MANAGED, m_pScreenVertBuf, nil);
  if FAILED(Result) then Exit;

  Result:= m_pd3dDevice.CreateVertexBuffer(7 * SizeOf(Single) * 4, 0,
      D3DFVF_XYZRHW or D3DFVF_TEX1 or D3DFVF_DIFFUSE, D3DPOOL_MANAGED, m_pScreenVertBuf2, nil);
  if FAILED(Result) then Exit;

  Result:= ComputeStepTexture;
  if FAILED(Result) then Exit;

  declVS[2] := D3DDECL_END;
  Result := m_pd3dDevice.CreateVertexDeclaration(@declVS, m_pFogDeclaration);
  if FAILED(Result) then Exit;

  declVS2[3] := D3DDECL_END;
  Result := m_pd3dDevice.CreateVertexDeclaration(@declVS2, m_pFogDeclaration2);
  if FAILED(Result) then Exit;

  Result:= D3DAPPERR_MEDIANOTFOUND;
  if FAILED(DXUtil_FindMediaFileCb(strPath, SizeOf(strPath), 'fogshader.vsh')) then Exit;
  Result := D3DXAssembleShaderFromFile(strPath, nil, nil, 0, @pCodeShader, nil);
  if FAILED(Result) then Exit;
  SAFE_RELEASE(m_pFogShader);
  Result := m_pd3dDevice.CreateVertexShader(PDWORD(pCodeShader.GetBufferPointer), m_pFogShader);
  if FAILED(Result) then Exit;
  SAFE_RELEASE(pCodeShader);

  Result:= D3DAPPERR_MEDIANOTFOUND;
  if FAILED(DXUtil_FindMediaFileCb(strPath, SizeOf(strPath), 'fogshader2.vsh')) then Exit;
  Result := D3DXAssembleShaderFromFile(strPath, nil, nil, 0, @pCodeShader, nil);
  if FAILED(Result) then Exit;
  SAFE_RELEASE(m_pFogShader2);
  Result := m_pd3dDevice.CreateVertexShader(PDWORD(pCodeShader.GetBufferPointer), m_pFogShader2);
  if FAILED(Result) then Exit;
  SAFE_RELEASE(pCodeShader);

  Result:= D3DAPPERR_MEDIANOTFOUND;
  if FAILED(DXUtil_FindMediaFileCb(strPath, SizeOf(strPath), 'FogPixShader.psh')) then Exit;
  Result := D3DXAssembleShaderFromFile(strPath, nil, nil, 0, @pCodeShader, nil);
  if FAILED(Result) then Exit;
  SAFE_RELEASE(m_pSubPicShader);
  Result := m_pd3dDevice.CreatePixelShader(PDWORD(pCodeShader.GetBufferPointer), m_pSubPicShader);
  if FAILED(Result) then Exit;
  SAFE_RELEASE(pCodeShader);

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.RestoreDeviceObjects: HResult;
type
  TVert = record
    x,y,z,w: Single;
    color: DWORD;
    u,v: Single;
  end;
function GetVerts(x,y,z,w: Single; color: DWORD; u,v: Single): TVert;
begin
  Result.x:= x;
  Result.y:= y;
  Result.z:= z;
  Result.w:= w;

  Result.color:= color;

  Result.u:= u;
  Result.v:= v;
end;
var
  s_Verts: array[0..3] of TVert;
  pVerts: Pointer;
begin
  m_pFont.RestoreDeviceObjects;
  m_pFontSmall.RestoreDeviceObjects;

  Result:= m_pd3dDevice.CreateTexture(RENDERTARGSIZE, RENDERTARGSIZE, 0,
      D3DUSAGE_RENDERTARGET, D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT, m_pBackFogDist, nil);
  if FAILED(Result) then Exit;

  Result:= m_pd3dDevice.CreateTexture(RENDERTARGSIZE, RENDERTARGSIZE, 0,
      D3DUSAGE_RENDERTARGET, D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT, m_pFrontFogDist, nil);
  if FAILED(Result) then Exit;

  Result:= m_pd3dDevice.CreateTexture(RENDERTARGSIZE, RENDERTARGSIZE, 0,
      D3DUSAGE_RENDERTARGET, D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT, m_pSceneDist, nil);
  if FAILED(Result) then Exit;

  Result:= m_pd3dDevice.CreateTexture(RENDERTARGSIZE, RENDERTARGSIZE, 0,
      D3DUSAGE_RENDERTARGET, D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT, m_pTempDist, nil);
  if FAILED(Result) then Exit;

  Result:= m_pd3dDevice.CreateDepthStencilSurface(RENDERTARGSIZE, RENDERTARGSIZE,
      D3DFMT_D16, D3DMULTISAMPLE_NONE, 0, True, m_pTargetZ, nil);
  if FAILED(Result) then Exit;

  Result:= m_pd3dDevice.GetRenderTarget(0, m_pOldRenderTarget);
  if FAILED(Result) then Exit;

  Result:= m_pd3dDevice.GetDepthStencilSurface(m_pOldStencilZ);
  if FAILED(Result) then Exit;

  // Set up a set of points which represents the screen
  s_Verts[0]:= GetVerts( 1.0,  1.0, 0.5, 1.0, D3DCOLOR_XRGB($ff, $ff, $8f), 1,0);
  s_Verts[1]:= GetVerts( 1.0, -1.0, 0.5, 1.0, D3DCOLOR_XRGB($ff, $ff, $8f), 1,1);
  s_Verts[2]:= GetVerts(-1.0,  1.0, 0.5, 1.0, D3DCOLOR_XRGB($ff, $ff, $8f), 0,0);
  s_Verts[3]:= GetVerts(-1.0, -1.0, 0.5, 1.0, D3DCOLOR_XRGB($ff, $ff, $8f), 0,1);

  Result:= m_pScreenVertBuf.Lock(0, SizeOf(s_Verts), Pointer(pVerts), 0);
  if FAILED(Result) then Exit;

  CopyMemory(pVerts, @s_Verts, SizeOf(s_Verts));

  Result:= m_pScreenVertBuf.Unlock;
  if FAILED(Result) then Exit;

  Result:= m_pScreenVertBuf2.Lock(0, SizeOf(s_Verts), Pointer(pVerts), 0);
  if FAILED(Result) then Exit;

  CopyMemory(pVerts, @s_Verts, SizeOf(s_Verts));

  Result:= m_pScreenVertBuf2.Unlock;
  if FAILED(Result) then Exit;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc: Called when the device-dependent objects are about to be lost.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InvalidateDeviceObjects: HResult;
begin
  m_pFont.InvalidateDeviceObjects;
  m_pFontSmall.InvalidateDeviceObjects;

  SAFE_RELEASE(m_pOldRenderTarget);
  SAFE_RELEASE(m_pOldStencilZ);
  SAFE_RELEASE(m_pFrontFogDist);
  SAFE_RELEASE(m_pBackFogDist);
  SAFE_RELEASE(m_pSceneDist);
  SAFE_RELEASE(m_pTempDist);
  SAFE_RELEASE(m_pTargetZ);

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.DeleteDeviceObjects: HResult;
begin
  m_pFont.DeleteDeviceObjects;
  m_pFontSmall.DeleteDeviceObjects;
  SAFE_RELEASE(m_pScreenVertBuf);
  SAFE_RELEASE(m_pScreenVertBuf2);
  SAFE_RELEASE(m_FogVolume.pVertexBuffer);
  SAFE_RELEASE(m_FogVolume.pIndexBuffer);
  SAFE_RELEASE(m_MeshShaderObject.pVertexBuffer);
  SAFE_RELEASE(m_MeshShaderObject.pIndexBuffer);
  SAFE_RELEASE(m_pMeshObject);
  SAFE_RELEASE(m_pStepTexture);
  SAFE_RELEASE(m_pSubPicShader);
  SAFE_RELEASE(m_pFogDeclaration);
  SAFE_RELEASE(m_pFogShader);
  SAFE_RELEASE(m_pFogDeclaration2);
  SAFE_RELEASE(m_pFogShader2);

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FinalCleanup: HResult;
begin
  SAFE_DELETE(m_pFont);
  SAFE_DELETE(m_pFontSmall);

  Result:= S_OK;
end;

end.

